/*     
 *  This file is part of DEW browser.
 *  Copyright (C) 2023 Woohyun Joh 
 *
 *  DEW browesr is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  DEW browesr is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use crate::content::Content;

use std::collections::HashMap;
use std::io::{Error, ErrorKind};
use std::mem;

pub struct MarkdownMarkupModel {
    pub string: String,
    pub content_idxs: Vec<usize>,
    pub content_map: HashMap<u32, Content>,
    pub content_keys: Vec<u32>,
}

impl TryFrom<&[u8]> for MarkdownMarkupModel {
    type Error = Error;
    fn try_from(slice: &[u8]) -> Result<MarkdownMarkupModel, Error> {
        enum ParsingState {
            NormalText,
            InnerValue,
            Key,
        }

        let source = String::from_utf8_lossy(slice).to_string();
        let mut content_map = HashMap::new();
        let mut content_keys = Vec::new();
        let mut string = String::new();
        let mut key = String::new();
        let mut inner_value = String::new();
        let mut content_idxs = Vec::new();

        let mut prev_char = char::default();
        let mut state = ParsingState::NormalText;
        for ch in source.chars() {
            if prev_char == '\\' {
                string.push(ch);
            } else {
                match state {
                    ParsingState::NormalText => match ch {
                        '[' => state = ParsingState::InnerValue,
                        _ => string.push(ch),
                    },
                    ParsingState::InnerValue => match ch {
                        '[' => return Err(Error::from(ErrorKind::InvalidData)),
                        ']' => state = ParsingState::Key,
                        _ => inner_value.push(ch),
                    },
                    ParsingState::Key => match ch {
                        '(' => return Err(Error::from(ErrorKind::InvalidData)),
                        ')' => {
                            state = ParsingState::NormalText;
                            let key_str = mem::take(&mut key);
                            let value_str = mem::take(&mut inner_value);
                            let first_char = match key_str.chars().nth(0) {
                                Some(c) => c,
                                None => return Err(Error::from(ErrorKind::InvalidData)),
                            };
                            let content = if first_char == '$' {
                                if value_str.chars().all(|c| c == '_') {
                                    Content::WritableText(value_str)
                                } else {
                                    Content::VariableText(value_str)
                                }
                            } else {
                                Content::Button(value_str)
                            };
                            let key_hash = crc32fast::hash(key_str.as_bytes());
                            content_map.insert(key_hash, content);
                            content_keys.push(key_hash);
                            content_idxs.push(string.len());
                        }
                        _ => key.push(ch),
                    },
                }
            }
            prev_char = ch;
        }

        Ok(MarkdownMarkupModel {
            string,
            content_idxs,
            content_map,
            content_keys,
        })
    }
}
