/*     
 *  This file is part of DEW browser.
 *  Copyright (C) 2023 Woohyun Joh 
 *
 *  DEW browesr is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  DEW browesr is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use document::Document;
use document::content::Content;
use script::context::Context;
use script::engine::Engine;
use script::ScriptError;

use std::collections::HashMap;

pub struct Tab {
    document: Document,
    cont_idx: usize,
    
    engine: Engine,
    script: HashMap<&'static str, Context>,
}

impl From<&[u8]> for Tab {
    fn from(slice: &[u8]) -> Tab {
        todo!()
    }
}

impl Tab {
    pub fn run_script(&mut self, script_name: &str) -> Result<(), ErrorFromTab> {
        match self.script.get(script_name) {
            Some(ctx) => Ok(self.engine.run(ctx)?),
            None => Err(ErrorFromTab::NoSuchScript),
        }
    }

    pub fn get_mut_content_by_key(&mut self, key: &str) -> Option<&mut Content> {
        let key_hash = crc32fast::hash(key.as_bytes());
        match self.document {
            Document::MMM(ref mut mmm) => mmm.content_map.get_mut(&key_hash),
            Document::III(_) => unimplemented!(),
        }
    }

    pub fn get_next_mut_content(&mut self) -> &mut Content {
        let next_idx = self.cont_idx + 1;
        match self.document {
            Document::MMM(ref mut mmm) => {
                let key_hash = match mmm.content_keys.iter().nth(next_idx) {
                    Some(key) => {
                        self.cont_idx = next_idx;
                        key
                    }
                    None => &mmm.content_keys[self.cont_idx],
                };
                mmm.content_map.get_mut(&key_hash).unwrap()
            }
            Document::III(_) => unimplemented!(),
        }
    }

    pub fn get_prev_mut_content(&mut self) -> &mut Content {
        let prev_idx = self.cont_idx - 1;
        match self.document {
            Document::MMM(ref mut mmm) => {
                let key_hash = match mmm.content_keys.iter().nth(prev_idx) {
                    Some(key) => {
                        self.cont_idx = prev_idx;
                        key
                    }
                    None => &mmm.content_keys[self.cont_idx],
                };
                mmm.content_map.get_mut(&key_hash).unwrap()
            }
            Document::III(_) => unimplemented!(),
        }
    }
}

pub enum ErrorFromTab {
    ScriptError(ScriptError),
    NoSuchScript,
}

impl From<ScriptError> for ErrorFromTab {
    fn from(e: ScriptError) -> ErrorFromTab {
        ErrorFromTab::ScriptError(e)
    }
}
