/*     
 *  This file is part of DEW browser.
 *  Copyright (C) 2023 Woohyun Joh 
 *
 *  DEW browesr is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  DEW browesr is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use crypto_box::{
    aead::{Aead, AeadCore, Payload},
    PublicKey, SalsaBox, SecretKey,
};
use rand_chacha::{rand_core::SeedableRng, ChaCha12Rng};

use std::io::{Error, ErrorKind};
use std::net::{SocketAddr, ToSocketAddrs, UdpSocket};

const KEY_SIZE: usize = 32;
const NONCE_SIZE: usize = 24;
const TAG_SIZE: usize = 16;
const SEED_SIZE: usize = 32;

const RELIABLE_FLAG: u8 = 0;
const UNRELIABLE_FLAG: u8 = 1;

///Rude User Datagram with Encryption
pub struct Rude {
    udp_socket: UdpSocket,
    cipher: SalsaBox,
    rng: ChaCha12Rng,
}

impl Rude {
    pub fn bind<A: ToSocketAddrs>(addr: A, seed: [u8; SEED_SIZE]) -> Result<Self, Error> {
        Ok(Self {
            udp_socket: UdpSocket::bind(addr)?,
            cipher: SalsaBox::new(
                &PublicKey::from([0; KEY_SIZE]),
                &SecretKey::from([0; KEY_SIZE]),
            ),
            rng: ChaCha12Rng::from_seed(seed),
        })
    }

    pub fn add_entropy(&mut self, entropy: impl AsRef<[u8]>) {
        let mut hasher = blake3::Hasher::new();
        hasher.update(&self.rng.get_seed());
        hasher.update(entropy.as_ref());

        let new_seed: [u8; SEED_SIZE] = *hasher.finalize().as_bytes();
        self.rng = ChaCha12Rng::from_seed(new_seed);
    }

    #[inline]
    pub fn local_addr(&self) -> Result<SocketAddr, Error> {
        self.udp_socket.local_addr()
    }

    #[inline]
    pub fn peer_addr(&self) -> Result<SocketAddr, Error> {
        self.udp_socket.peer_addr()
    }

    pub fn accept(&mut self) -> Result<(), Error> {
        let secret_key = SecretKey::generate(&mut self.rng);

        let mut public_key = [0u8; KEY_SIZE];
        let (_len, src) = self.udp_socket.recv_from(&mut public_key)?;

        self.udp_socket.connect(src)?;
        self.udp_socket
            .send(&secret_key.public_key().as_bytes().clone())?;

        self.cipher = SalsaBox::new(&PublicKey::from(public_key), &secret_key);

        Ok(())
    }

    pub fn connect<A: ToSocketAddrs>(&mut self, addr: A) -> Result<(), Error> {
        let secret_key = SecretKey::generate(&mut self.rng);
        self.udp_socket.connect(addr)?;
        self.udp_socket
            .send(&secret_key.public_key().as_bytes().clone())?;

        let mut public_key = [0u8; KEY_SIZE];
        self.udp_socket.recv(&mut public_key)?;

        self.cipher = SalsaBox::new(&PublicKey::from(public_key), &secret_key);

        Ok(())
    }

    pub fn send_reliably(&mut self, buf: &[u8], aad: impl AsRef<[u8]>) -> Result<(), Error> {
        let mut buf_w_flag = buf.to_vec();
        buf_w_flag.push(RELIABLE_FLAG);

        let nonce = SalsaBox::generate_nonce(&mut self.rng);
        let ciphertext = self
            .cipher
            .encrypt(
                &nonce,
                Payload {
                    msg: buf_w_flag.as_slice(),
                    aad: aad.as_ref(),
                },
            )
            .unwrap();

        let mut trans_buf = nonce.to_vec();
        trans_buf.extend(ciphertext.clone());

        self.udp_socket
            .set_read_timeout(Some(std::time::Duration::from_millis(100)))?;
        loop {
            self.udp_socket.send(&trans_buf)?;

            let mut reply = vec![0u8; buf_w_flag.len() + TAG_SIZE];
            match self.udp_socket.recv(&mut reply) {
                Err(e) if e.kind() == ErrorKind::TimedOut => continue,
                Err(e) => return Err(e),
                Ok(_) => {
                    if reply == ciphertext {
                        return Ok(());
                    } else {
                        return Err(Error::from(ErrorKind::InvalidData));
                    }
                }
            }
        }
    }

    pub fn send_unreliably(&mut self, buf: &[u8], aad: impl AsRef<[u8]>) -> Result<(), Error> {
        let mut buf_w_flag = buf.to_vec();
        buf_w_flag.push(UNRELIABLE_FLAG);

        let nonce = SalsaBox::generate_nonce(&mut self.rng);
        let ciphertext = self
            .cipher
            .encrypt(
                &nonce,
                Payload {
                    msg: buf_w_flag.as_slice(),
                    aad: aad.as_ref(),
                },
            )
            .unwrap();

        let mut trans_buf = nonce.to_vec();
        trans_buf.extend(ciphertext);

        self.udp_socket.send(&trans_buf)?;
        Ok(())
    }

    pub fn recv(&self, buf: &mut [u8], aad: impl AsRef<[u8]>) -> Result<bool, Error> {
        let mut trans_buf = vec![0u8; NONCE_SIZE + buf.len() + 1 + TAG_SIZE];
        self.udp_socket.recv(&mut trans_buf)?;

        let (nonce, ciphertext) = trans_buf.split_at(NONCE_SIZE);
        let plaintext = self
            .cipher
            .decrypt(
                nonce.into(),
                Payload {
                    msg: ciphertext,
                    aad: aad.as_ref(),
                },
            )
            .unwrap();

        let flag = *plaintext.last().unwrap();
        if flag == RELIABLE_FLAG {
            self.udp_socket.send(ciphertext)?;
        }

        buf.iter_mut()
            .zip(plaintext.iter())
            .for_each(|(b, p)| *b = *p);

        Ok(flag == RELIABLE_FLAG)
    }
}
