/*     
 *  This file is part of DEW browser.
 *  Copyright (C) 2023 Woohyun Joh 
 *
 *  DEW browesr is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  DEW browesr is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

pub mod data;
pub mod value;

use crate::data::Data;
use crate::value::Value;

use std::collections::HashMap;
use std::mem;

pub struct PersonalDataBase {
    map: HashMap<u32, Data>,
}

impl From<&[u8]> for PersonalDataBase {
    fn from(slice: &[u8]) -> PersonalDataBase {
        let mut map = HashMap::new();
        let mut buf = slice;
        while buf.len() > 0 {
            let (key_slice, latter) = buf.split_at(mem::size_of::<u32>());
            let key_hash = u32::from_le_bytes(key_slice.try_into().unwrap());
            let (data, size) = match latter[0] {
                b'B' => (
                    Data::Public(Value::Bool(latter[1] != 0)),
                    mem::size_of::<bool>(),
                ),
                b'b' => (
                    Data::Private(Value::Bool(latter[1] != 0)),
                    mem::size_of::<bool>(),
                ),
                b'S' => (
                    Data::Public(Value::Sint(i64::from_le_bytes(
                        latter[1..1 + mem::size_of::<i64>()].try_into().unwrap(),
                    ))),
                    mem::size_of::<i64>(),
                ),
                b's' => (
                    Data::Private(Value::Sint(i64::from_le_bytes(
                        latter[1..1 + mem::size_of::<i64>()].try_into().unwrap(),
                    ))),
                    mem::size_of::<i64>(),
                ),
                b'R' => (
                    Data::Public(Value::Real(f64::from_le_bytes(
                        latter[1..1 + mem::size_of::<f64>()].try_into().unwrap(),
                    ))),
                    mem::size_of::<f64>(),
                ),
                b'r' => (
                    Data::Private(Value::Real(f64::from_le_bytes(
                        latter[1..1 + mem::size_of::<f64>()].try_into().unwrap(),
                    ))),
                    mem::size_of::<f64>(),
                ),
                b'T' => {
                    let bytes_until_null: Vec<u8>;
                    (
                        Data::Public(Value::Text({
                            bytes_until_null = latter
                                .iter()
                                .cloned()
                                .take_while(|&byte| byte != 0)
                                .collect();
                            String::from_utf8_lossy(&bytes_until_null).to_string()
                        })),
                        bytes_until_null.len(),
                    )
                }
                b't' => {
                    let bytes_until_null: Vec<u8>;
                    (
                        Data::Private(Value::Text({
                            bytes_until_null = latter
                                .iter()
                                .cloned()
                                .take_while(|&byte| byte != 0)
                                .collect();
                            String::from_utf8_lossy(&bytes_until_null).to_string()
                        })),
                        bytes_until_null.len(),
                    )
                }
                _ => panic!("Invalid Byte"),
            };

            buf = &latter[1 + size..];
            map.insert(key_hash, data);
        }

        PersonalDataBase { map }
    }
}

impl PersonalDataBase {
    #[inline]
    pub fn get_value(&self, key: &str) -> Option<Value> {
        let key_hash = crc32fast::hash(key.as_bytes());
        Some(match self.map.get(&key_hash)? {
            Data::Public(v) => v.clone(),
            Data::Private(v) => v.clone(),
        })
    }

    #[inline]
    pub fn set_value(&mut self, key: &str, value: Value) -> Option<Value> {
        let key_hash = crc32fast::hash(key.as_bytes());
        Some(match self.map.get_mut(&key_hash)? {
            Data::Public(v) => mem::replace(v, value),
            Data::Private(v) => mem::replace(v, value),
        })
    }

    #[inline]
    pub fn get_data(&self, key: &str) -> Option<Data> {
        let key_hash = crc32fast::hash(key.as_bytes());
        self.map.get(&key_hash).cloned()
    }

    #[inline]
    pub fn set_data(&mut self, key: &str, data: Data) -> Option<Data> {
        let key_hash = crc32fast::hash(key.as_bytes());
        Some(mem::replace(self.map.get_mut(&key_hash)?, data))
    }
}
