/*     
 *  This file is part of DEW browser.
 *  Copyright (C) 2023 Woohyun Joh 
 *
 *  DEW browesr is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  DEW browesr is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use crate::array::Array;
use crate::value::Value;

use std::mem;

pub struct Context {
    pub bytecode: Vec<u16>,
    pub immediates: Array<Value>,
    pub func_ptrs: Vec<usize>,

    pub var_names: Vec<String>,
    pub func_names: Vec<String>,
}

impl From<&[u8]> for Context {
    fn from(slice: &[u8]) -> Context {
        let (bytecode_len, latter) = slice.split_at(mem::size_of::<u32>());
        let (immediates_len, latter) = latter.split_at(mem::size_of::<u32>());
        let (func_ptrs_len, latter) = latter.split_at(mem::size_of::<u32>());
        let (var_names_len, latter) = latter.split_at(mem::size_of::<u32>());
        let (func_names_len, latter) = latter.split_at(mem::size_of::<u32>());

        let (bytecode_slice, latter) =
            latter.split_at(u32::from_le_bytes(bytecode_len.try_into().unwrap()) as usize);
        let (immediates_slice, latter) =
            latter.split_at(u32::from_le_bytes(immediates_len.try_into().unwrap()) as usize);
        let (func_ptrs_slice, latter) =
            latter.split_at(u32::from_le_bytes(func_ptrs_len.try_into().unwrap()) as usize);
        let (var_names_slice, latter) =
            latter.split_at(u32::from_le_bytes(var_names_len.try_into().unwrap()) as usize);
        let (func_names_slice, latter) =
            latter.split_at(u32::from_le_bytes(func_names_len.try_into().unwrap()) as usize);

        assert!(latter.is_empty());

        Context {
            bytecode: bytecode_slice
                .chunks(mem::size_of::<u16>())
                .map(|slice| u16::from_le_bytes(slice.try_into().unwrap()))
                .collect(),
            immediates: {
                let mut buffer = immediates_slice;
                let mut immediates = Array::default();
                while !buffer.is_empty() {
                    let (value, size) = match buffer[0] {
                        b'B' => (Value::Bool(buffer[1] != 0), mem::size_of::<bool>()),
                        b'S' => (
                            Value::Sint(i64::from_le_bytes(
                                buffer[1..1 + mem::size_of::<i64>()].try_into().unwrap(),
                            )),
                            mem::size_of::<i64>(),
                        ),
                        b'R' => (
                            Value::Real(f64::from_le_bytes(
                                buffer[1..1 + mem::size_of::<f64>()].try_into().unwrap(),
                            )),
                            mem::size_of::<f64>(),
                        ),
                        b'T' => {
                            let bytes_until_null: Vec<u8>;
                            (
                                Value::Text({
                                    bytes_until_null = latter
                                        .iter()
                                        .cloned()
                                        .take_while(|&byte| byte != 0)
                                        .collect();
                                    String::from_utf8_lossy(&bytes_until_null).to_string()
                                }),
                                bytes_until_null.len(),
                            )
                        }
                        _ => panic!("Invaild Byte format"),
                    };
                    buffer = &buffer[1 + size..];
                    immediates.push(value);
                }

                immediates
            },
            func_ptrs: func_ptrs_slice
                .chunks(mem::size_of::<u32>())
                .map(|slice| u32::from_le_bytes(slice.try_into().unwrap()) as usize)
                .collect(),

            var_names: var_names_slice
                .split(|&byte| byte == b'\0')
                .filter(|segment| !segment.is_empty())
                .map(|segment| String::from_utf8_lossy(segment).to_string())
                .collect(),
            func_names: func_names_slice
                .split(|&byte| byte == b'\0')
                .filter(|segment| !segment.is_empty())
                .map(|segment| String::from_utf8_lossy(segment).to_string())
                .collect(),
        }
    }
}
