/*     
 *  This file is part of DEW browser.
 *  Copyright (C) 2023 Woohyun Joh 
 *
 *  DEW browesr is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  DEW browesr is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use crate::array::Array;
use crate::context::Context;
use crate::opcode::*;
use crate::stack::Stack;
use crate::value::Value;
use crate::ScriptError;

use std::ops::*;

macro_rules! arithmetic {
    ($args:expr, $op:ident) => {{
        let left = $args.pop()?;
        let right = $args.pop()?;
        match (left, right) {
            (Value::Sint(l), Value::Sint(r)) => Ok(Value::Sint(l.$op(r))),
            (Value::Real(l), Value::Real(r)) => Ok(Value::Real(l.$op(r))),
            (Value::Real(_), Value::Sint(_)) => Err(ScriptError::TypeMismatch),
            (Value::Sint(_), Value::Real(_)) => Err(ScriptError::TypeMismatch),
            _ => Err(ScriptError::InvalidOperation),
        }
    }};
}

macro_rules! bitwise {
    ($args:expr, $op:ident) => {{
        let left = $args.pop()?;
        let right = $args.pop()?;
        match (left, right) {
            (Value::Bool(l), Value::Bool(r)) => Ok(Value::Bool(l.$op(r))),
            (Value::Sint(l), Value::Sint(r)) => Ok(Value::Sint(l.$op(r))),
            (Value::Bool(_), Value::Sint(_)) => Err(ScriptError::TypeMismatch),
            (Value::Sint(_), Value::Bool(_)) => Err(ScriptError::TypeMismatch),
            _ => Err(ScriptError::InvalidOperation),
        }
    }};
}

macro_rules! shift {
    ($args:expr, $op:ident) => {{
        let left = $args.pop()?;
        let right = $args.pop()?;
        match (left, right) {
            (Value::Sint(l), Value::Sint(r)) => Ok(Value::Sint(l.$op(r))),
            _ => Err(ScriptError::InvalidOperation),
        }
    }};
}

#[derive(Default)]
pub struct Engine {
    ext_func: Vec<Box<dyn Fn(Stack<Value>) -> Result<Value, ScriptError>>>,
}

impl Engine {
    pub fn new() -> Engine {
        let mut engine = Engine::default();

        engine.register_fn(|mut args| arithmetic!(args, add));
        engine.register_fn(|mut args| arithmetic!(args, sub));
        engine.register_fn(|mut args| arithmetic!(args, mul));
        engine.register_fn(|mut args| arithmetic!(args, div));
        engine.register_fn(|mut args| arithmetic!(args, rem));

        engine.register_fn(|mut args| bitwise!(args, bitand));
        engine.register_fn(|mut args| bitwise!(args, bitor));
        engine.register_fn(|mut args| bitwise!(args, bitxor));

        engine.register_fn(|mut args| shift!(args, shl));
        engine.register_fn(|mut args| shift!(args, shr));

        engine.register_fn(|mut args| {
            if let Value::Sint(s) = args.pop()? {
                Ok(Value::Sint(!s))
            } else {
                Err(ScriptError::InvalidOperation)
            }
        });

        //TODO add type castings and comparings

        engine
    }

    #[inline]
    pub fn register_fn(
        &mut self,
        func: impl Fn(Stack<Value>) -> Result<Value, ScriptError> + 'static,
    ) {
        self.ext_func.push(Box::new(func))
    }

    pub fn run(&self, context: &Context) -> Result<(), ScriptError> {
        let mut memory = Array::default();
        let mut stack = Stack::default();
        let mut call_stack = Vec::default();
        let mut loop_stack = Stack::default();
        let mut life_stack = Stack::default();
        let mut pc = 0;
        while pc < context.bytecode.len() {
            let opcode = context.bytecode[pc] >> OPERAND_SIZE;
            let operand = (context.bytecode[pc] & (u16::MAX >> OPCODE_SIZE)) as usize;
            match opcode {
                NEW => match operand {
                    0 => memory.push(stack.pop()?),
                    _ => memory.push(Value::List(Array::from(vec![Value::default(); operand]))),
                },
                POP => memory.index_mut(operand)?.assign(stack.pop()?)?,
                IMD => stack.push(context.immediates.index(operand)?.clone()),
                VAR => stack.push(memory.index(operand)?.clone()),
                EFN => {
                    let arg_stack = std::mem::take(&mut stack);
                    stack.push(self.ext_func[operand](arg_stack)?);
                }
                IFN => {
                    call_stack.push(pc);
                    pc = context.func_ptrs[operand];
                }
                RET => match call_stack.pop() {
                    Some(prev) => pc = prev,
                    None => return Ok(()),
                },
                GET => {
                    let idx = stack.pop()?.expect_sint()? as usize;
                    stack.push(memory.index(operand)?.expect_list()?.index(idx)?.clone());
                }
                SET => {
                    let idx = stack.pop()?.expect_sint()? as usize;
                    let value = stack.pop()?;
                    memory
                        .index(operand)?
                        .expect_list()?
                        .index_mut(idx)?
                        .assign(value)?;
                }

                IF_ => {
                    if stack.pop()?.expect_bool()? {
                        life_stack.push(memory.len());
                    } else {
                        pc += operand;
                    }
                }
                LPS => {
                    loop_stack.push((pc, pc + operand));
                    life_stack.push(memory.len());
                }
                BRK => {
                    loop_stack.truncate(loop_stack.len() - operand);
                    let (_begin, end) = loop_stack.pop()?;
                    pc = end;
                }
                NXT => {
                    loop_stack.truncate(loop_stack.len() - operand);
                    let (begin, _end) = loop_stack.pop()?;
                    pc = begin;
                }
                END => {
                    //TODO test this code
                    life_stack.truncate(operand);
                    memory.truncate(life_stack.pop()?);

                    let (begin, end) = loop_stack.last()?;
                    if pc == *end {
                        pc = *begin;
                    }
                }

                _ => return Err(ScriptError::InvalidOpcode),
            }

            pc += 1;
        }

        Ok(())
    }
}
