/*     
 *  This file is part of DEW browser.
 *  Copyright (C) 2023 Woohyun Joh 
 *
 *  DEW browesr is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  DEW browesr is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use pdb::value::Value as PDBValue;

use crate::array::Array;
use crate::ScriptError;

#[derive(Clone, PartialEq)]
pub enum Value {
    Void,
    Bool(bool),
    Sint(i64),
    Real(f64),
    Text(String),
    List(Array<Value>),
}

impl Default for Value {
    fn default() -> Value {
        Value::Void
    }
}

impl From<PDBValue> for Value {
    fn from(pdb_value: PDBValue) -> Value {
        match pdb_value {
            PDBValue::Bool(b) => Value::Bool(b),
            PDBValue::Sint(s) => Value::Sint(s),
            PDBValue::Real(r) => Value::Real(r),
            PDBValue::Text(t) => Value::Text(t),
        }
    }
}

impl Value {
    pub fn assign(&mut self, other: Value) -> Result<(), ScriptError> {
        if *self == Value::Void {
            *self = other;
            return Ok(());
        }

        match (self, other) {
            (Value::Bool(l), Value::Bool(r)) => Ok(*l = r),
            (Value::Sint(l), Value::Sint(r)) => Ok(*l = r),
            (Value::Real(l), Value::Real(r)) => Ok(*l = r),
            (Value::Text(l), Value::Text(r)) => Ok(*l = r),
            (Value::List(l), Value::List(r)) => Ok(*l = r),
            _ => Err(ScriptError::TypeMismatch),
        }
    }

    pub fn expect_sint(&self) -> Result<i64, ScriptError> {
        match self {
            Value::Sint(v) => Ok(v.clone()),
            _ => Err(ScriptError::NotExpectedType),
        }
    }

    pub fn expect_list(&self) -> Result<Array<Value>, ScriptError> {
        match self {
            Value::List(v) => Ok(v.clone()),
            _ => Err(ScriptError::NotExpectedType),
        }
    }

    pub fn expect_bool(&self) -> Result<bool, ScriptError> {
        match self {
            Value::Bool(v) => Ok(v.clone()),
            _ => Err(ScriptError::NotExpectedType),
        }
    }
}
